/*
  ALCATEL SpeedTouch USB modem microcode upload & ADSL link UP utility
  Copyright (C) 2001 Benoit PAPILLAULT
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author   : Benoit PAPILLAULT <benoit.papillault@free.fr>
  Creation : 05/03/2001
  
  This program uploads the microcode to the ALCATEL SpeedTouch USB modem.
  It uses libusb (http://libusb.sourceforge.net/).
  
  The microcode can be uploaded only once. If the upload is correct,
  ADSL led should be both green & red.
  
  If you try to download the microcode twice, you will get tons of
  timeout errors.

  25/03/2001: added GPL license.

  13/04/2001: removed timeout messages.
*/

/* for ident(1) command */
static char id[] = "@(#) $Id: modem_run.c,v 1.10 2001/04/25 21:41:46 papillau Exp $";

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include "usbfs.h"
#include "modem.h"
#include "extract.h"

#define CTRL_TIMEOUT 1000
#define DATA_TIMEOUT 1000

void print_char(unsigned char c)
{
  if (c>=' ' && c<0x7f)
    printf("%c",c);
  /*
    else if (c=='\r' || c=='\n' || c=='\t' || c=='\b')
    printf("%c",c);
  */
  else	printf(".");
}

void dump(unsigned char *buf,int len)
{
  int i,j;
  
  for (i=0;i<len;i+=16)
    {
      for (j=i;j<len && j<i+16;j++)
	printf("%02x ",buf[j]);
      for (;j<i+16;j++)
	printf("   ");
      for (j=i;j<len && j<i+16;j++)
	print_char(buf[j]);
      printf("\n");
    }
}

unsigned char data1[] = {
  0x88, 0xf8, 0x00, 0x00, 0x00, 0x00, 0xf2, 0x01, 
  0x0d, 0x00, 0x00, 0xea, 0x0c, 0x00, 0x00, 0xea, 
  0x0b, 0x00, 0x00, 0xea, 0x0a, 0x00, 0x00, 0xea, 
  0x09, 0x00, 0x00, 0xea, 0x08, 0x00, 0x00, 0xea, 
  0xa4, 0x00, 0x00, 0xea, 0x06, 0x00, 0x00, 0xea, 
  0x32, 0x11, 0xa0, 0xe3, 0x5c, 0x03, 0x9f, 0xe5, 
  0x00, 0x00, 0x81, 0xe5, 0x58, 0x13, 0x9f, 0xe5, 
  0x00, 0x00, 0xa0, 0xe3, 0x00, 0x00, 0x81, 0xe5, 
  0xf8, 0xff, 0xff, 0xea, 0x01, 0x00, 0xa0, 0xe3, 
  0x48, 0x13, 0x9f, 0xe5, 0x00, 0x00, 0x81, 0xe5, 
  0x3c, 0x13, 0x9f, 0xe5, 0x04, 0x00, 0xa0, 0xe3, 
  0x00, 0x00, 0x81, 0xe5, 0xd1, 0x00, 0xa0, 0xe3, 
  0x00, 0xf0, 0x29, 0xe1, 0x01, 0xd9, 0xa0, 0xe3, 
  0xd2, 0x00, 0xa0, 0xe3, 0x00, 0xf0, 0x29, 0xe1, 
  0x24, 0xd3, 0x9f, 0xe5, 0xd3, 0x00, 0xa0, 0xe3, 
  0x00, 0xf0, 0x29, 0xe1, 0x1c, 0xd3, 0x9f, 0xe5, 
  0x00, 0x00, 0xa0, 0xe3, 0x00, 0x10, 0xa0, 0xe3, 
  0xb0, 0x02, 0x9f, 0xe5, 0x08, 0x00, 0x80, 0xe2, 
  0xbc, 0x12, 0x9f, 0xe5, 0x00, 0x10, 0x80, 0xe5, 
  0x5f, 0x10, 0xe0, 0xe3, 0x5d, 0x30, 0xe0, 0xe3, 
  0x5c, 0x50, 0xe0, 0xe3, 0x5b, 0x70, 0xe0, 0xe3, 
  0xac, 0x02, 0x9f, 0xe5, 0xa4, 0x22, 0x9f, 0xe5, 
  0xa8, 0x42, 0x9f, 0xe5, 0xa8, 0x62, 0x9f, 0xe5, 
  0x00, 0x10, 0x80, 0xe5, 0x00, 0x30, 0x82, 0xe5, 
  0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
  0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
  0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
  0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
  0x00, 0x70, 0x86, 0xe5, 0x54, 0x02, 0x9f, 0xe5, 
  0x08, 0x00, 0x80, 0xe2, 0x5c, 0x12, 0x9f, 0xe5, 
  0x00, 0x10, 0x80, 0xe5, 0x6c, 0x02, 0x9f, 0xe5, 
  0x6c, 0x12, 0x9f, 0xe5, 0x00, 0x10, 0x80, 0xe5, 
  0x00, 0x10, 0xa0, 0xe3, 0x98, 0x02, 0x9f, 0xe5, 
  0x00, 0x10, 0x80, 0xe5, 0x94, 0x02, 0x9f, 0xe5, 
  0x00, 0x10, 0x80, 0xe5, 0x01, 0x09, 0xa0, 0xe3, 
  0x00, 0x10, 0xa0, 0xe3, 0x02, 0x26, 0xa0, 0xe3, 
  0x04, 0x30, 0xa0, 0xe3, 0x03, 0x10, 0x80, 0xe6, 
  0x02, 0x00, 0x50, 0xe1, 0xfc, 0xff, 0xff, 0x1a, 
  0x7f, 0x19, 0xa0, 0xe3, 0x1e, 0x30, 0xa0, 0xe3, 
  0x03, 0x3b, 0x83, 0xe2, 0x00, 0x00, 0xa0, 0xe3, 
  0x00, 0x2a, 0x83, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
  0x14, 0x00, 0x50, 0xe3, 0x04, 0x20, 0x81, 0xe4, 
  0xfa, 0xff, 0xff, 0xba, 0x02, 0xe0, 0xa0, 0xe3, 
  0x03, 0xeb, 0x8e, 0xe2, 0x0e, 0x20, 0xa0, 0xe1, 
  0x18, 0x00, 0x50, 0xe3, 0x04, 0x00, 0x00, 0xaa, 
  0x00, 0xca, 0x82, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
  0x18, 0x00, 0x50, 0xe3, 0x04, 0xc0, 0x81, 0xe4, 
  0xfa, 0xff, 0xff, 0xba, 0x40, 0x00, 0x50, 0xe3, 
  0x04, 0x00, 0x00, 0xaa, 0x00, 0x2a, 0x83, 0xe1, 
  0x01, 0x00, 0x80, 0xe2, 0x40, 0x00, 0x50, 0xe3, 
  0x04, 0x20, 0x81, 0xe4, 0xfa, 0xff, 0xff, 0xba, 
  0x01, 0x0a, 0x50, 0xe3, 0x04, 0x00, 0x00, 0xaa, 
  0x00, 0x2a, 0x8e, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
  0x01, 0x0a, 0x50, 0xe3, 0x04, 0x20, 0x81, 0xe4, 
  0xfa, 0xff, 0xff, 0xba, 0x7f, 0x09, 0xa0, 0xe3, 
  0x03, 0x40, 0xa0, 0xe3, 0x10, 0x0f, 0x02, 0xee, 
  0x10, 0x4f, 0x03, 0xee, 0x10, 0x0f, 0x07, 0xee, 
  0x10, 0x0f, 0x05, 0xee, 0x3d, 0x00, 0xa0, 0xe3, 
  0x10, 0x0f, 0x01, 0xee, 0x01, 0x73, 0xa0, 0xe3, 
  0x04, 0x00, 0xa0, 0xe3, 0x05, 0x00, 0xc7, 0xe5, 
  0x01, 0x00, 0xa0, 0xe3, 0x06, 0x00, 0xc7, 0xe5, 
  0x01, 0x00, 0xa0, 0xe3, 0x07, 0x00, 0xc7, 0xe5, 
  0x03, 0x00, 0xa0, 0xe3, 0x08, 0x00, 0xc7, 0xe5, 
  0x00, 0x00, 0x40, 0x01, 0x12, 0x88, 0xd4, 0xf2, 
  0x01, 0x00, 0x00, 0xce, 0x01, 0xa0, 0xe3, 0x09, 
  0x00, 0xc7, 0xe5, 0x80, 0x00, 0xa0, 0xe3, 0x01, 
  0x00, 0xc7, 0xe5, 0x53, 0x00, 0xa0, 0xe3, 0x00, 
  0xf0, 0x29, 0xe1, 0x94, 0x11, 0x9f, 0xe5, 0x21, 
  0x03, 0xa0, 0xe3, 0x00, 0x10, 0x80, 0xe5, 0x10, 
  0x10, 0xa0, 0xe3, 0x22, 0x01, 0xa0, 0xe3, 0x00, 
  0x10, 0x80, 0xe5, 0x04, 0x00, 0x8f, 0xe2, 0x01, 
  0x00, 0x80, 0xe2, 0x10, 0xff, 0x2f, 0xe1, 0x5e, 
  0x4f, 0xb8, 0x78, 0x40, 0x1e, 0xfc, 0xd1, 0x38, 
  0x78, 0xf9, 0x7a, 0x09, 0x02, 0xbc, 0x7a, 0x21, 
  0x43, 0x09, 0x02, 0x7c, 0x7a, 0x21, 0x43, 0x09, 
  0x02, 0x3c, 0x7a, 0x21, 0x43, 0x11, 0x38, 0x01, 
  0x28, 0x2d, 0xdb, 0x03, 0x28, 0x2b, 0xdc, 0x80, 
  0x00, 0x55, 0x4c, 0x24, 0x58, 0xa7, 0x46, 0x46, 
  0x4b, 0x0e, 0x33, 0x7a, 0x7b, 0x12, 0x02, 0x3c, 
  0x7b, 0x22, 0x43, 0x12, 0xd0, 0x14, 0x1c, 0xf4, 
  0x3c, 0xff, 0x3c, 0x0e, 0xd5, 0x40, 0x07, 0x02, 
  0xd0, 0x1c, 0x1c, 0x0b, 0x1c, 0x21, 0x1c, 0x52, 
  0x1e, 0x8c, 0x5c, 0x9c, 0x54, 0xfb, 0xd1, 0x3b, 
  0xa4, 0x14, 0x23, 0x23, 0x60, 0xc0, 0x0f, 0x64, 
  0x30, 0x00, 0xe0, 0x66, 0x20, 0x78, 0x70, 0xcb, 
  0xe7, 0x01, 0xa2, 0x10, 0x47, 0x00, 0x00, 0x01, 
  0xf0, 0xa0, 0xe1, 0x04, 0x20, 0x8f, 0xe2, 0x01, 
  0x20, 0x82, 0xe2, 0x12, 0xff, 0x2f, 0xe1, 0x60, 
  0x20, 0xf0, 0xe7, 0x04, 0xe0, 0x4e, 0xe2, 0x00, 
  0x40, 0x2d, 0xe9, 0x00, 0xe0, 0x4f, 0xe1, 0x07, 
  0x40, 0x2d, 0xe9, 0xa8, 0x10, 0x8f, 0xe2, 0x00, 
  0x00, 0x91, 0xe5, 0x01, 0x00, 0x20, 0xe2, 0x00, 
  0x00, 0x81, 0xe5, 0x00, 0x00, 0x50, 0xe3, 0x03, 
  0x00, 0x00, 0x0a, 0x0d, 0x10, 0xa0, 0xe3, 0xcc, 
  0x20, 0x9f, 0xe5, 0x00, 0x10, 0x82, 0xe5, 0x02, 
  0x00, 0x00, 0xea, 0x0f, 0x10, 0xa0, 0xe3, 0xbc, 
  0x20, 0x9f, 0xe5, 0x00, 0x10, 0x82, 0xe5, 0x01, 
  0x00, 0xa0, 0xe3, 0x90, 0x10, 0x9f, 0xe5, 0x00, 
  0x00, 0x81, 0xe5, 0x6c, 0x10, 0x8f, 0xe2, 0x00, 
  0x00, 0x91, 0xe5, 0x01, 0x00, 0x40, 0xe2, 0x00, 
  0x00, 0x81, 0xe5, 0x00, 0x00, 0x50, 0xe3, 0x02, 
  0x00, 0x00, 0x1a, 0x94, 0x00, 0x9f, 0xe5, 0x94, 
  0x10, 0x9f, 0xe5, 0x00, 0x00, 0x81, 0xe5, 0x90, 
  0x20, 0x9f, 0xe5, 0x00, 0x10, 0x92, 0xe5, 0x07, 
  0x40, 0xbd, 0xe8, 0x0e, 0xf0, 0x69, 0xe1, 0x00, 
  0x80, 0xfd, 0xe8, 0x00, 0x00, 0x00, 0x8c, 0x20, 
  0x00, 0x00, 0x88, 0x50, 0x00, 0x00, 0x01, 0x02, 
  0x00, 0x00, 0x00, 0xf4, 0x00, 0x00, 0x00, 0x39, 
  0x00, 0x00, 0x00, 0x10, 0x40, 0x08, 0x00, 0x00, 
  0x40, 0x00, 0x00, 0x20, 0x40, 0x00, 0x00, 0x30, 
  0x4e, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa0, 0x15, 
  0x80, 0x00, 0x00, 0x15, 0x2f, 0x8c, 0x9b, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x5d, 
  0x02, 0x00, 0x00, 0x5d, 0x02, 0x00, 0x00, 0x97, 
  0x02, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x40, 
  0x00, 0x40, 0x01, 0x64, 0x00, 0x50, 0x01, 0x70, 
  0x3e, 0x00, 0x00, 0xe0, 0x3c, 0x00, 0x00, 0x20, 
  0x00, 0x00, 0x90, 0x20, 0x00, 0x00, 0xa4, 0x85, 
  0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x78, 
  0x03, 0x00, 0x00, 0x58, 0x00, 0x50, 0x01, 0x70, 
  0xc3, 0x00, 0x00, 0x60, 0x00, 0x50, 0x01, 0x04, 
  0x00, 0x00, 0x84, 0x40, 0x01, 0x12, 0x08, 0x04, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x14
};

void test_int1_alt2(int fdusb)
{
  const int interface = 1;
  const int alternate = 2;

  if (usbfs_claimintf(fdusb,interface) < 0)
    perror("usb_claimintf");
  else
    {
      if (usbfs_set_interface(fdusb,interface,alternate) < 0)
	perror("usbfs_set_interface");

      if (usbfs_releaseintf(fdusb,interface) < 0)
	perror("usbfs_releaseintf");
    }
}

int my_usb_clear(int fdusb, int ep)
{
  int n;
  
  n = usbfs_control_msg(fdusb,
			0x02, /*USB_RECIP_ENDPOINT,*/
			0x01, /*USB_REQ_CLEAR_FEATURE,*/
			0,ep,NULL,0,CTRL_TIMEOUT);
  if (n < 0)
    perror("CLEAR_FEATURE");

  n = usbfs_resetep(fdusb,ep);
  if (n < 0)
    perror("usb_resetep");

  return 0;
}

void usage()
{
  printf("modem_run version $Name: tar_gz_26042001 $\n");
  printf("usage: modem_run [-v] [-f microcode] [-m]\n");
  printf("  -v : verbose mode\n");
  printf("  -f microcode : upload this microcode file first\n");
  printf("  -m : monitor the line\n");
  printf("  --help : this message\n");
  exit (-1);
}

/* returns -1 in case of error */

int upload_microcode(int fdusb, const char * filename,int verbose)
{
	int n, ret;
	struct timeval start, stop;
	unsigned char * uc_buf, buf[0x1000];
	/* buf should be at least 511 bytes */
	int uc_len;

	uc_buf = extract_microcode(filename,&uc_len);
	if (uc_buf == NULL)
	{
		printf("microcode can not be located... :-(\n");
		return -1;
	}
	
	if ((ret=usbfs_claimintf(fdusb,2)) <0)
    {
		perror("usbfs_claimintf 2");
		return ret;
    }
	
	if (usbfs_clear_halt(fdusb,EP_CODE_OUT) < 0)
		perror("usbfs_clear_halt EP_CODE_OUT");
	
	if (usbfs_clear_halt(fdusb,EP_CODE_IN) < 0)
		perror("usbfs_clear_halt EP_CODE_IN");
	
	/* URB 8 : both leds are static green */
	if (verbose) gettimeofday(&start,NULL);
	n = usbfs_bulk_msg(fdusb,EP_CODE_OUT,data1,sizeof(data1),100);
	if (verbose) gettimeofday(&stop,NULL);
	if (n < 0)
    {
		perror("usbfs_bulk_msg [w]");
		usbfs_releaseintf(fdusb,2);
		return n;
    }
	else
    {
		printf("BLOCK1 : %6d bytes   uploaded : OK",sizeof(data1));
		if (verbose)
			printf(" [%f ms]",(double)(stop.tv_sec -start.tv_sec) * 1000.0
				   + (double)(stop.tv_usec - start.tv_usec) / 1000.0);
		printf("\n");
    }
	
	/* after usb_bulk_write :
	   + USB  led : blinking green
	   + ADSL led : off
	*/
	
	/* URB 9 & 10 */
	/*my_usb_clear(fdusb,0x85);*/
	
	/* URB 11 */
	if (verbose) gettimeofday(&start,NULL);
	n = usbfs_bulk_msg(fdusb,EP_CODE_IN,buf,0x200,DATA_TIMEOUT);
	if (verbose) gettimeofday(&stop,NULL);
	if (n < 0)
    {
		perror("usb_bulk_msg [r]");
		usbfs_releaseintf(fdusb,2);
		return n;
    }
	else
    {
		printf("BLOCK2 : %6d bytes downloaded : OK",n);
		if (verbose)
			printf(" [%f ms]",(double)(stop.tv_sec -start.tv_sec) * 1000.0
				   + (double)(stop.tv_usec - start.tv_usec) / 1000.0);
		printf("\n");
		if (verbose) dump(buf,n);
    }
	
	
	/* URB 12 to 139 */
	
	if (verbose) gettimeofday(&start,NULL);
	n = usbfs_bulk_msg(fdusb,EP_CODE_OUT,uc_buf,uc_len,DATA_TIMEOUT);
	if (verbose) gettimeofday(&stop,NULL);
	if (n < 0)
	{
		perror("usb_bulk_msg [w]");
		usbfs_releaseintf(fdusb,2);
		return n;
	}
	else
	{
		printf("BLOCK3 : %6d bytes   uploaded : OK",uc_len);
		if (verbose)
			printf(" [%f ms]",(double)(stop.tv_sec -start.tv_sec) * 1000.0
				   + (double)(stop.tv_usec - start.tv_usec) / 1000.0);
		printf("\n");
	}

	/* before downloading microcode :
	   + USB  led : blinking
	   + ADSL led : off
	   
	   during downloading:
	   + USB  led : blinking green
	   + ADSL led : off
	   
	   after downloading microcode :
	   + USB  led : static green
	   + ADSL led : static red
	*/
	
	/* URB 140 & 141 */
	/*my_usb_clear(fdusb,0x85);*/
	
	/* URB 142 */
	if (verbose) gettimeofday(&start,NULL);
	n = usbfs_bulk_msg(fdusb,EP_CODE_IN,buf,0x200,DATA_TIMEOUT);
	if (verbose) gettimeofday(&stop,NULL);
	if (n < 0)
    {
		perror("usb_bulk_msg [r]");
		usbfs_releaseintf(fdusb,2);
		return n;
    }
	else
    {
		printf("BLOCK4 : %6d bytes downloaded : OK",n);
		if (verbose)
			printf(" [%f ms]",(double)(stop.tv_sec -start.tv_sec) * 1000.0
				   + (double)(stop.tv_usec - start.tv_usec) / 1000.0);
		printf("\n");
		if (verbose) dump(buf,n);
    }
	
	if (usbfs_releaseintf(fdusb,2) < 0)
		perror("usbfs_releaseintf 2");

  return 0;
}

int get_reference(int fdusb)
{
  char buf[0xe];
  int i,n;
  
  /* URB 151 : this buffer match (if you reverse one byte with its neighbour)
   the label on the manta box!!! in my case : 3EC18607CAAB06 */

  /* original line */
  n = usbfs_control_msg(fdusb,0xc0,0x26,0xf7,0,buf,sizeof(buf),CTRL_TIMEOUT);
  if (n<0)
    perror("get_reference: usbfs_control_msg");
  else
    {
      printf("Reference     : ");
      //dump(buf,n);
      for (i=0;i<n;i++)
	{
	  int idx = ((i%2)==0) ? i+1 : i-1;
	  if (i==3 || i==12)
	    printf(" ");
	  print_char(buf[idx]);
	}
      printf(" (see under your modem box)\n");
    }

  return n;
}

int my_usb_control_msg(int fdusb,int request,int value,int size)
{
    unsigned char buf[100];
    int n;

    if (size > sizeof(buf))
      {
	printf("my_usb_control_msg: buffer is too small\n");
	return -1;
      }

    n = usbfs_control_msg(fdusb,0xc0,request,value,0x00,buf,size,100);
    if (n < 0)
      {
	perror("my_usb_control_msg: usbfs_control_msg");
	return -1;
      }
    /*
    printf("c0 %02x %04x : size %d ",request,value,n);
    dump(buf,n);
    */
    return 0;
}

#define OFFSET_7  0 /* size 1 */
#define OFFSET_b  1 /* size 8 */
#define OFFSET_d  9 /* size 4 */
#define OFFSET_e 13 /* size 1 */
#define OFFSET_f 14 /* size 1 */
#define TOTAL    15

#define SIZE_7 1
#define SIZE_b 8
#define SIZE_d 4
#define SIZE_e 1
#define SIZE_f 1

/* buf MUST BE at least 15 bytes */

int get_state(int fdusb, unsigned char *buf)
{
  if (usbfs_bulk_msg(fdusb,0x81,buf,0x10,100) != -1)
    printf("gotcha!\n"); /* always fail? */

  memset(buf,0,TOTAL);
  usbfs_control_msg(fdusb,0xc0,0x12,0x07,0x00,buf+OFFSET_7,SIZE_7,100);
  usbfs_control_msg(fdusb,0xc0,0x12,0x0b,0x00,buf+OFFSET_b,SIZE_b,100);
  usbfs_control_msg(fdusb,0xc0,0x12,0x0d,0x00,buf+OFFSET_d,SIZE_d,100);
  usbfs_control_msg(fdusb,0xc0,0x01,0x0e,0x00,buf+OFFSET_e,SIZE_e,100);
  usbfs_control_msg(fdusb,0xc0,0x01,0x0f,0x00,buf+OFFSET_f,SIZE_f,100);

  return 0;
}

/*
  This a magic sequence, totaly unknown to me, it comes from the Windows
  driver ... 
*/

void test_sequence(int fdusb)
{
  unsigned char buf[10];
  int n;

  /* URB 147 */
  buf[0] = 0x1c; buf[1] = 0x50;
  n = usbfs_control_msg(fdusb,0x40,0x01,0x0b,0x00,buf,2,100);
  if (n<0)
    perror("URB147");

  /* URB 148 */
  buf[0] = 0x32; buf[1] = 0x00;
  n = usbfs_control_msg(fdusb,0x40,0x01,0x02,0x00,buf,2,100);
  if (n<0)
    perror("URB148");

  /* URB 149 */
  buf[0] = 0x01; buf[1] = 0x00; buf[2] = 0x01;
  n = usbfs_control_msg(fdusb,0x40,0x01,0x03,0x00,buf,3,100);
  if (n<0)
    perror("URB149");

  /* URB 150 */
  buf[0] = 0x01; buf[1] = 0x00; buf[2] = 0x01;
  n = usbfs_control_msg(fdusb,0x40,0x01,0x04,0x00,buf,3,100);
  if (n<0)
    perror("URB150");
}

/* buf MUST be 15 bytes long */

int print_state(unsigned char *buf, int verbose)
{
  time_t t = time(NULL);
  int down_speed, up_speed;

  if (verbose)
    {
      printf("=== modem state at %.24s ===\n",ctime(&t));
      printf("c0 12 0007 : "); dump(buf+OFFSET_7,SIZE_7);
      printf("c0 12 000b : "); dump(buf+OFFSET_b,SIZE_b);
      printf("c0 12 000d : "); dump(buf+OFFSET_d,SIZE_d);
      printf("c0 01 000e : "); dump(buf+OFFSET_e,SIZE_e);
      printf("c0 01 000f : "); dump(buf+OFFSET_f,SIZE_f);
    }
      
  switch (buf[OFFSET_7])
    {
    case 0x00:
      printf("ADSL line is down\n");
      break;
    case 0x08:
      printf("ADSL line is blocked? \n");
      break;
    case 0x10:
      printf("ADSL line is synchronising\n");
      break;
    case 0x20:

      down_speed = buf[OFFSET_b] | (buf[OFFSET_b+1]<<8)
	| (buf[OFFSET_b+2]<<16) | (buf[OFFSET_b+3]<<24);
      up_speed = buf[OFFSET_b+4] | (buf[OFFSET_b+5]<<8)
	| (buf[OFFSET_b+6]<<16) | (buf[OFFSET_b+7]<<24);

      printf("ADSL line is up, downstream at %d kbit/s, "
	     "upstream at %d kbit/s\n",down_speed,up_speed);
      break;
    }
  
  return 0;
}

/*
  send an ioctl code to the speedtouch.c kernel module.
  It is handled by usbdevfs, see the following funtion in
  /usr/src/linux/drivers/usb/devio.c : usbdev_ioctl() & proc_ioctl()
  /usr/src/linux/drivers/usb/speedtouch.c : udsl_usb_ioctl()
*/

int my_usb_ioctl(int fdusb,int cmd)
{
  struct usbdevfs_ioctl s;
  int ret;

  /* speedtouch.c manage interface 1 */
  s.ifno = 1; 

  /* should be 1 for UDSL_IOCTL_START and 2 for UDSL_IOCTL_STOP */
  s.ioctl_code = cmd;
  s.data = NULL;

  if ((ret=usbfs_ioctl(fdusb,&s)) < 0)
    {
      perror("ioctl");
      printf("maybe your forget to do 'insmod speedtch'\n");
    }

  return ret;
}

int main(int argc, char *argv[])
{
	int fdusb;
	int i;
	int verbose = 0, monitoring = 0;
	unsigned char buf[TOTAL+1000], prevbuf[TOTAL];
	const char * file = NULL;
	
	for (i=1;i<argc;i++)
		if (strcmp(argv[i],"-v")==0)
			verbose = 1;
		else if (strcmp(argv[i],"-m")==0)
			monitoring = 1;
		else if (strcmp(argv[i],"-f")==0 && i+1<argc)
			file = argv[++i];
		else if (strcmp(argv[i],"--help")==0)
			usage();
		else
			usage();
	
	if (file == NULL)
		usage();
	
	fdusb = usbfs_search("/proc/bus/usb",ST_VENDOR,ST_PRODUCT);
	if (fdusb < 0)
    {
		printf("I'm sorry, I didn't find your ADSL modem!\n");
		printf("Tips: check that you are running this program as root"
			   " and your device is seen by Linux"
			   " by checking /proc/bus/usb/devices\n");
		return -1;
    }
	
	printf("Gotcha, I found your ADSL ALCATEL SpeedTouch USB modem!\n");
	
	if (usbfs_set_configuration(fdusb,1) < 0)
    {
		perror("usbfs_set_configuration 1");
		return -1;
    }
	
	if (file != NULL)
    {
		if (upload_microcode (fdusb,file,verbose) < 0)
		{
			printf("Uploading %s failed! Continuing anyway ...\n",file);
		}
		
		sleep(1);
    }
	
	if (usbfs_set_configuration(fdusb,1) < 0)
    {
		perror("usbfs_set_configuration 1");
		return -1;
    }
	
	test_int1_alt2(fdusb);
	
	if (usbfs_claimintf(fdusb,0) <0)
    {
		perror("usbfs_claimintf 0");
		return -1;
    }
	
	get_reference(fdusb);
	
	/* some 'reset' on endpoint */
	if (usbfs_clear_halt(fdusb,0x81) < 0)
		perror("usbfs_clear_halt");
	
	test_sequence(fdusb);
	
	/* should start line sync */
	my_usb_control_msg(fdusb,0x12,0x04,2);
	
	get_state(fdusb,prevbuf);
	print_state(prevbuf,verbose);
	
	if (monitoring)
    {
		for (;;)
		{
			/* wait a second before getting the new modem state,
			   it avoids to consume too much CPU */
			sleep(1);
			
			get_state(fdusb,buf);
			if (memcmp(buf,prevbuf,TOTAL) != 0)
			{
				print_state(buf,verbose);
				
				if ( (buf[OFFSET_7] == 0x20 && prevbuf[OFFSET_7] != 0x20))
				{
					printf("ADSL link goes UP\n");
					sleep(2);
					break;
				}
				
				if (buf[OFFSET_7] != 0x20
					&& prevbuf[OFFSET_7] == 0x20)
				{
					printf("ADSL link goes DOWN\n");
				}
				
				memcpy(prevbuf,buf,TOTAL);
			}
		}
    }
	
	if (usbfs_releaseintf(fdusb,0) < 0)
		perror("usb_releaseintf");
	
	close(fdusb);
	return 0;
}

