/*
  ALCATEL SpeedTouch USB modem utility : PPPoA implementation (2nd edition)
  Copyright (C) 2001 Benoit PAPILLAULT
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author : Benoit PAPILLAULT <benoit.papillault@free.fr>
  Creation : 21/03/2001

  This program is designed to work under pppd, with the option "pty".

  21/03/2001: you MUST run 'modem_init' and 'modem_link' before trying to
  run this one.

  25/03/2001: 'modem_link' has been replaced by 'modem_diag'.

  28/03/2001 : removed 'test_main'. replace libusb by usbfs. 

  01/04/2001 : removed reading cells in signal handler.

  12/04/2001: added handling of HDLC line discipline. In fact, pppd
  sends us HDLC packet in stdin (stdin is normally the master side of
  a master/slave tty). However, with no special care, HDLC packets can
  be concatenated and since HDLC header does not include any length
  indicator, HDLC is not handled the way it should be.

  After reading the source code from rp-pppoe, it tried to do :
    ioctl(0,TIOCSETD,&N_HDLC).
  It works if your kernel has been compiled with HDLC support
  (maybe insmod n_hdlc).
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <termios.h> /* N_HDLC & TIOCSETD */
#include <sys/resource.h> /* setpriority() */

/* my USB library */
#include "usbfs.h"
#include "pool.h"

/* for ident(1) command */
static char id[] = "@(#) $Id: pppoa2.c,v 1.24 2001/04/25 22:37:24 papillau Exp $";

/* USB level */
#define DATA_TIMEOUT 1000

#include "modem.h"

/* ATM level : adapt according your ADSL provider settings */
int my_vpi = -1;
int my_vci = -1;

#define CELL_SIZE 53
#define CELL_HDR  05
#define CELL_DATA 48

/* two pools : one if for sending URB, the other
   for sending buffers */

struct pool urb_pool;
struct pool buf_pool;

/* PPP level : DO NOT MODIFY */
#define syncHDLC 1

/* global parameters */

int debug = 0;
int usb_sync = 1;
pid_t child_pid;

/* GDB level : define to 1 if you want to debug with GDB */
/*#define GDB_DEBUG*/

#define CRC32_REMAINDER CBF43926
#define CRC32_INITIAL 0xffffffff

#define CRC32(c,crc) (crc32tab[((size_t)(crc>>24) ^ (c)) & 0xff] ^ (((crc) << 8)))

unsigned long  crc32tab[256] =
{
 0x00000000L, 0x04C11DB7L, 0x09823B6EL, 0x0D4326D9L,
 0x130476DCL, 0x17C56B6BL, 0x1A864DB2L, 0x1E475005L,
 0x2608EDB8L, 0x22C9F00FL, 0x2F8AD6D6L, 0x2B4BCB61L,
 0x350C9B64L, 0x31CD86D3L, 0x3C8EA00AL, 0x384FBDBDL,
 0x4C11DB70L, 0x48D0C6C7L, 0x4593E01EL, 0x4152FDA9L,
 0x5F15ADACL, 0x5BD4B01BL, 0x569796C2L, 0x52568B75L,
 0x6A1936C8L, 0x6ED82B7FL, 0x639B0DA6L, 0x675A1011L,
 0x791D4014L, 0x7DDC5DA3L, 0x709F7B7AL, 0x745E66CDL,
 0x9823B6E0L, 0x9CE2AB57L, 0x91A18D8EL, 0x95609039L,
 0x8B27C03CL, 0x8FE6DD8BL, 0x82A5FB52L, 0x8664E6E5L,
 0xBE2B5B58L, 0xBAEA46EFL, 0xB7A96036L, 0xB3687D81L,
 0xAD2F2D84L, 0xA9EE3033L, 0xA4AD16EAL, 0xA06C0B5DL,
 0xD4326D90L, 0xD0F37027L, 0xDDB056FEL, 0xD9714B49L,
 0xC7361B4CL, 0xC3F706FBL, 0xCEB42022L, 0xCA753D95L,
 0xF23A8028L, 0xF6FB9D9FL, 0xFBB8BB46L, 0xFF79A6F1L,
 0xE13EF6F4L, 0xE5FFEB43L, 0xE8BCCD9AL, 0xEC7DD02DL,
 0x34867077L, 0x30476DC0L, 0x3D044B19L, 0x39C556AEL,
 0x278206ABL, 0x23431B1CL, 0x2E003DC5L, 0x2AC12072L,
 0x128E9DCFL, 0x164F8078L, 0x1B0CA6A1L, 0x1FCDBB16L,
 0x018AEB13L, 0x054BF6A4L, 0x0808D07DL, 0x0CC9CDCAL,
 0x7897AB07L, 0x7C56B6B0L, 0x71159069L, 0x75D48DDEL,
 0x6B93DDDBL, 0x6F52C06CL, 0x6211E6B5L, 0x66D0FB02L,
 0x5E9F46BFL, 0x5A5E5B08L, 0x571D7DD1L, 0x53DC6066L,
 0x4D9B3063L, 0x495A2DD4L, 0x44190B0DL, 0x40D816BAL,
 0xACA5C697L, 0xA864DB20L, 0xA527FDF9L, 0xA1E6E04EL,
 0xBFA1B04BL, 0xBB60ADFCL, 0xB6238B25L, 0xB2E29692L,
 0x8AAD2B2FL, 0x8E6C3698L, 0x832F1041L, 0x87EE0DF6L,
 0x99A95DF3L, 0x9D684044L, 0x902B669DL, 0x94EA7B2AL,
 0xE0B41DE7L, 0xE4750050L, 0xE9362689L, 0xEDF73B3EL,
 0xF3B06B3BL, 0xF771768CL, 0xFA325055L, 0xFEF34DE2L,
 0xC6BCF05FL, 0xC27DEDE8L, 0xCF3ECB31L, 0xCBFFD686L,
 0xD5B88683L, 0xD1799B34L, 0xDC3ABDEDL, 0xD8FBA05AL,
 0x690CE0EEL, 0x6DCDFD59L, 0x608EDB80L, 0x644FC637L,
 0x7A089632L, 0x7EC98B85L, 0x738AAD5CL, 0x774BB0EBL,
 0x4F040D56L, 0x4BC510E1L, 0x46863638L, 0x42472B8FL,
 0x5C007B8AL, 0x58C1663DL, 0x558240E4L, 0x51435D53L,
 0x251D3B9EL, 0x21DC2629L, 0x2C9F00F0L, 0x285E1D47L,
 0x36194D42L, 0x32D850F5L, 0x3F9B762CL, 0x3B5A6B9BL,
 0x0315D626L, 0x07D4CB91L, 0x0A97ED48L, 0x0E56F0FFL,
 0x1011A0FAL, 0x14D0BD4DL, 0x19939B94L, 0x1D528623L,
 0xF12F560EL, 0xF5EE4BB9L, 0xF8AD6D60L, 0xFC6C70D7L,
 0xE22B20D2L, 0xE6EA3D65L, 0xEBA91BBCL, 0xEF68060BL,
 0xD727BBB6L, 0xD3E6A601L, 0xDEA580D8L, 0xDA649D6FL,
 0xC423CD6AL, 0xC0E2D0DDL, 0xCDA1F604L, 0xC960EBB3L,
 0xBD3E8D7EL, 0xB9FF90C9L, 0xB4BCB610L, 0xB07DABA7L,
 0xAE3AFBA2L, 0xAAFBE615L, 0xA7B8C0CCL, 0xA379DD7BL,
 0x9B3660C6L, 0x9FF77D71L, 0x92B45BA8L, 0x9675461FL,
 0x8832161AL, 0x8CF30BADL, 0x81B02D74L, 0x857130C3L,
 0x5D8A9099L, 0x594B8D2EL, 0x5408ABF7L, 0x50C9B640L,
 0x4E8EE645L, 0x4A4FFBF2L, 0x470CDD2BL, 0x43CDC09CL,
 0x7B827D21L, 0x7F436096L, 0x7200464FL, 0x76C15BF8L,
 0x68860BFDL, 0x6C47164AL, 0x61043093L, 0x65C52D24L,
 0x119B4BE9L, 0x155A565EL, 0x18197087L, 0x1CD86D30L,
 0x029F3D35L, 0x065E2082L, 0x0B1D065BL, 0x0FDC1BECL,
 0x3793A651L, 0x3352BBE6L, 0x3E119D3FL, 0x3AD08088L,
 0x2497D08DL, 0x2056CD3AL, 0x2D15EBE3L, 0x29D4F654L,
 0xC5A92679L, 0xC1683BCEL, 0xCC2B1D17L, 0xC8EA00A0L,
 0xD6AD50A5L, 0xD26C4D12L, 0xDF2F6BCBL, 0xDBEE767CL,
 0xE3A1CBC1L, 0xE760D676L, 0xEA23F0AFL, 0xEEE2ED18L,
 0xF0A5BD1DL, 0xF464A0AAL, 0xF9278673L, 0xFDE69BC4L,
 0x89B8FD09L, 0x8D79E0BEL, 0x803AC667L, 0x84FBDBD0L,
 0x9ABC8BD5L, 0x9E7D9662L, 0x933EB0BBL, 0x97FFAD0CL,
 0xAFB010B1L, 0xAB710D06L, 0xA6322BDFL, 0xA2F33668L,
 0xBCB4666DL, 0xB8757BDAL, 0xB5365D03L, 0xB1F740B4L 
};
                                                                
unsigned int calc_crc(unsigned char *mem, int len, unsigned int initial)
{
  unsigned int crc;
  crc = initial;

  for(;len;mem++,len--)
    {
      crc = CRC32(*mem, crc);
    }
  return(crc);
}

void print_char(unsigned char c)
{
  if (c>=' ' && c<0x7f)
    printf("%c",c);
  /*
    else if (c=='\r' || c=='\n' || c=='\t' || c=='\b')
    printf("%c",c);
  */
  else  printf(".");
}

void dump(unsigned char *buf,int len)
{
  int i,j;

  for (i=0;i<len;i+=16)
    {
      for (j=i;j<len && j<i+16;j++)
        printf("%02x ",buf[j]);
      for (;j<i+16;j++)
        printf("   ");
      for (j=i;j<len && j<i+16;j++)
        print_char(buf[j]);
      printf("\n");
    }
}

/*
  ppp_read returns a pointer to a static buffer.
*/

int ppp_read(int fd,unsigned char ** buf, int *n)
{
  /* we must handle HDLC framing, since this is what pppd
     sends to us. I use some code from rp-pppoe
  */

  static unsigned char ppp_buf[64*1024];
  int r;


  for (;;)
    {
      r = read(fd,ppp_buf,sizeof(ppp_buf));
      /*printf("ppp_read : r = %d, %02x %02x\n",r,ppp_buf[0],ppp_buf[1]);*/

      if (r<0)
	{
	  if (errno == EINTR)
	    continue; 
	  
	  perror("error reading from pppd");
	  return r;
	}
      
      /* supress leading 2 bytes */
      
      if (syncHDLC)
	{
	  if (r<3)
	    {
	      printf("ppp_read : short on data, r=%d => ignored\n",r);
	      continue;
	    }

	  if (r==sizeof(ppp_buf))
	    {
	      printf("ppp_read : too much data\n");
	      continue;
	    }

	  *buf = ppp_buf + 2;
	  *n   = r - 2;

	  return (r-2);
	}
      else
	{
	  printf("ppp_read: async HDLC not handled\n");
	}

      break;
    }
}

int ppp_write(int fd,unsigned char * buf, int n)
{
  unsigned char ppp_buf[64*1024];
  int r;

  if (syncHDLC)
    {
      ppp_buf[0] = 0xff; //FRAME_ADDR;
      ppp_buf[1] = 0x03; //FRAME_CTRL;
      
      memcpy(ppp_buf +2, buf, n);
      if ((r=write(fd,ppp_buf,n+2)) < 0)
	perror("error writing to pppd");
      return r;
    }
  else
    {
      printf("ppp_write: async HDLC not handled\n");
    }

  return -1;
}

/* 
   parse an ATM cells and try to complete (or make a new one)
   AAL5 frame. If succes, the frame is sent to pppd. In case of
   error, the current UNIX process is terminated (may be, pppd
   already exited).
*/

int aal5_read(unsigned char * cell_buf,size_t count,
	      unsigned int vpi,unsigned int vci,unsigned int pti,
	      int fdout)
{
  static unsigned char buf[64*1024];
  static int len = 0;
  
  int ret;
  unsigned int real_len;

  if (len == 0)
    memset(buf,0,sizeof(buf));
  
  if (vpi != my_vpi || vci != my_vci)
    {
      printf("vpi.vci mismatch ... ignored!\n");
      return 0;
    }
      
  if (len + CELL_DATA < sizeof(buf))
    {
      /*      printf("copying cell at offset %d\n",len);*/
      memcpy(buf+len, cell_buf, CELL_DATA);
      len += CELL_DATA;
    }
  else
    printf("aal5_read: buffer overflow\n");

  /* we should add a check for OAM cells */
  if ((pti & 2) == 0)
    return 0;

  real_len = (buf[len -6] << 8) | buf[len -5];

  if (debug)
    printf("aal5_read: len=%d => real_len=%d\n",len,real_len);

  if ((ret=ppp_write(fdout,buf,real_len)) < 0)
      return ret;

  len = 0;
  
  return 0;
}

/*
  cell_read : is used to decode an individual cell and to try to make
  an AAL5 frame.
*/

int cell_read(unsigned char *cellbuf, size_t count, int fdout)
{
  unsigned int vpi, vci, pti;

  /* decoding & printing of the cell */
  vpi = (cellbuf[0] << 4) | (cellbuf[1]>>4);
  vci = ((cellbuf[1] & 0xf) <<12) | (cellbuf[2]<<4) | (cellbuf[3]>>4);
  pti = (cellbuf[3] & 0xf);

  /* filling our AAL5 frame */
  return aal5_read(cellbuf+CELL_HDR,CELL_DATA,vpi,vci,pti,fdout);
}

int cell_write(unsigned char * cellbuf,
	       unsigned char *buf, int n,
	       unsigned int vpi, unsigned int vci, unsigned int pti)
{
  unsigned char gfc = 0;
  unsigned char hec = 0xec;

  if (n != CELL_DATA)
    return -1;

  /* normalize */

  vpi &= 0xff;
  vci &= 0xffff;
  pti &= 0xf;

  cellbuf[0] = (gfc<<4) | (vpi>>4);
  cellbuf[1] = (vpi<<4) | (vci>>12);
  cellbuf[2] = (vci>>4);
  cellbuf[3] = ((vci&0xf) << 4) | pti;
  cellbuf[4] = hec;

  memcpy(cellbuf+CELL_HDR,buf,CELL_DATA);
  return CELL_DATA;
}

/* Warning: (ok, tooks me half an hour to find my own buf)
   buf must be large enought to contains the whole AAL5 frame,
   with its layer.
*/

int aal5_write(int fdusb, unsigned char *buf, int n)
{
  /* add the trailer */
  int total_len = 48* ((n+8+CELL_DATA-1)/48);
  unsigned int crc, pti = 0;

  /* this buffer will contain all cells for an AAL5 frame,
     so AAL5 frame is max 64K + 48 that's 1367 cells at max
  */
  unsigned char * bigbuf;
  int ptr;
  struct timeval tv;
  struct buffer * tbuf;
  int ret;

  /* get a buffer from the pool */

  if (!buffer_pool_alloc(&buf_pool,&tbuf))
    {
      printf("buf_alloc failed : loosing data\n");
      return -1;
    }

  bigbuf = (unsigned char *)tbuf->buf;

  memset(buf+n,0,total_len-n);

  buf[total_len -6] = n>>8;
  buf[total_len -5] = n&0xff;

  crc = ~calc_crc(buf,total_len-4,-1);

  buf[total_len -4] = crc >> 24;
  buf[total_len -3] = crc >> 16;
  buf[total_len -2] = crc >>  8;
  buf[total_len -1] = crc;

  ptr = 0;
  while (total_len>0)
    {
      if (total_len <= CELL_DATA)
	/* this is the last cell, so mark it */
	pti = 2;

      cell_write(bigbuf+ptr,buf,CELL_DATA,my_vpi,my_vci,pti);
      ptr += CELL_SIZE;

      buf       += CELL_DATA;
      total_len -= CELL_DATA;
    }

  if (debug)
    {
      gettimeofday(&tv,NULL);
      printf("[%d.%06d] writing %d bytes\n",tv.tv_sec,tv.tv_usec,ptr);
      dump(bigbuf,ptr);
    }
  
  if (usb_sync)
    {
      if ((ret=usbfs_bulk_msg(fdusb,EP_DATA_OUT,bigbuf,ptr,DATA_TIMEOUT)) < 0)
	perror("error writing to usb");

      buffer_pool_free(&buf_pool,tbuf);
    }
  else
    {
      /* usbfs_bulk_msg() can take up to 150ms : so we use submit_urb() */
      
      struct buffer * urb_buf;
      struct usbdevfs_urb * mon_urb;
      
      if (!buffer_pool_alloc(&urb_pool,&urb_buf))
	printf("urb_alloc : failed\n");
      
      mon_urb = (struct usbdevfs_urb *) urb_buf -> buf;
      
      usbfs_fill_bulk_urb(mon_urb,EP_DATA_OUT,bigbuf,ptr,SIGRTMIN);
      
      if ((ret=usbfs_submit_urb(fdusb,mon_urb)) < 0)
	perror("usbfs_submit_urb");
    }

  return ret;
}

/* variable that will be used in several function
   => so global */

struct usbdevfs_urb gurb_int;
unsigned char       gbuf[100];
int                 gfdusb, gfdout;

void sig_int(struct usbdevfs_urb * urb)
{
  usbfs_fill_bulk_urb(&gurb_int,EP_INT,gbuf,sizeof(gbuf),SIGRTMIN);
  
  if (usbfs_submit_urb(gfdusb,&gurb_int) < 0)
    perror("usbfs_submit_urb");
}

#define URB_NB 10
#define URB_BUF_SIZE 64*53

static unsigned char buffer[URB_NB][URB_BUF_SIZE];
static struct usbdevfs_urb gurb[URB_NB];
static int urb_submitted[URB_NB];

int submit_unsubmitted_urb(int fdusb)
{
  int i, ret;

  for (i=0;i<URB_NB;i++)
    if (!urb_submitted[i])
      {
	/* to be sure we does not get previous data : zeroed the buffer */
	memset(buffer[i],0,sizeof(buffer[i]));
	usbfs_fill_bulk_urb(&gurb[i],EP_DATA_IN,
			    buffer[i],sizeof(buffer[i]),SIGRTMIN);
	
	/* mark the urb submitted before, to avoid re-submitting it
	   in the sig handler */
	urb_submitted[i] = 1;

	if ((ret=usbfs_submit_urb(fdusb,&gurb[i])) < 0)
	  {
	    urb_submitted[i] = 0;
	    perror("oh no! usbfs_submit_urb");
	    return ret;
	  }
      }
  
  return 0;
}

void sig_data_out(struct usbdevfs_urb * urb)
{
  /* we free buffers we allocated before */

  struct buffer * tbuf;

  tbuf = buffer_pool_search(&buf_pool,urb->buffer);
  if (tbuf != NULL)
    buffer_pool_free(&buf_pool,tbuf);
  else
    printf("can't free buf : %p\n",urb->buffer);

  tbuf = buffer_pool_search(&urb_pool,urb);
  if (tbuf != NULL)
    buffer_pool_free(&urb_pool,tbuf);
  else
    printf("can't free urb : %p\n",urb);

}

int decode_usb_pkt(unsigned char *buf, int len)
{
  int r;

  /* ... ??? ... I prefer to say nothing about the following lines */
  
  if ((len % CELL_SIZE) != 0)
    {
      int junk = len % CELL_SIZE;
      printf("junk len%%%d = %d\n",CELL_SIZE,junk);
      len -= junk;
      buf += junk;
    }

  while (len >= CELL_SIZE)
    {
      if ((r = cell_read(buf,CELL_SIZE,gfdout)) < 0)
	  return r;

      buf += CELL_SIZE;
      len -= CELL_SIZE;
    };

  if (len != 0)
    printf("argh : len = %d\n",len);

  return 0;
}

void sig_data_in(struct usbdevfs_urb * urb)
{
  int i, urbi;
  unsigned char * buf;
  int len;
  struct timeval tv;

  for (i=0;i<URB_NB;i++)
    if (urb == &gurb[i])
      {
	urbi = i;
	if (debug)
	  {
	    gettimeofday(&tv,NULL);
	    printf("[%d.%06d] reading %d bytes\n",
		   tv.tv_sec, tv.tv_usec,urb->actual_length);
	  }
	break;
      }

  buf = (unsigned char *)urb->buffer;
  len = urb->actual_length;

  if (debug)
    dump(buf,len);

  decode_usb_pkt(buf,len);

  urb_submitted[urbi] = 0;
  submit_unsubmitted_urb(gfdusb);
}

void mon_signal(int sig, siginfo_t * info, void *ptr)
{
  //  struct usbdevfs_urb * urb = (struct usbdevfs_urb *)info->si_addr;
  struct usbdevfs_urb * urbb;
  struct timeval tv;
  int idx;

  /* if we choose to handle synchronous read/write, we does nothing here */
  if (usb_sync)
    return ;

  if (debug)
    {
      gettimeofday(&tv,NULL);
      printf("%d.%06d : mon_signal\n",tv.tv_sec,tv.tv_usec);
    }

  /*
  printf("si_signo = %d\n",info->si_signo);
  printf("si_errno = %d\n",info->si_errno);
  printf("si_code  = %d\n",info->si_code);
  printf("si_addr  = %p\n",info->si_addr);

  printf("urb = %p\n",urb);
  */

  for (idx=0;;idx++)
    {
      if (usbfs_reapurb_ndelay(gfdusb,&urbb) < 0)
	{
	  if (errno != EAGAIN)
	    perror("usbfs_reapurb");

	  break;
	}

      if (debug)
	{
	  gettimeofday(&tv,NULL);
	  printf("%d.%06d : urb=%p, buffer %p, buffer_length %d,"
		 " actual_length=%d, status=%d, error_count=%d"
		 " [%d]\n",
		 tv.tv_sec,tv.tv_usec,
		 urbb,urbb->buffer,urbb->buffer_length,
		 urbb->actual_length,urbb->status,urbb->error_count,
		 idx);
	}

      switch (urbb->endpoint)
	{
	case EP_INT:
	  sig_int(urbb);
	  break;
	  
	case EP_DATA_IN:
	  sig_data_in(urbb);
	  break;

	case EP_DATA_OUT:
	  sig_data_out(urbb);
	  break;
	  
	default:
	  printf("unhandled endpoint 0x%x\n",urbb->endpoint);
	  printf("urb.buffer_length = %d\n",urbb->buffer_length);
	  printf("urb.actual_length = %d\n",urbb->actual_length);
	  printf("urb.status = %d\n",urbb->status);
	  printf("urb.error_count = %d\n",urbb->error_count);

	  break;
	}
    }
}

/* endpoint 0x81, which is on interface 0 is used for handling 'interrupt'.
   We will handle it that way : setup an urb , submit it,
   in the signal handler, reap the urb and submit a new one.
   and so on ... 
*/
 
void handle_endpoint_81(int fd)
{
  struct sigaction act;

  /* we 'clean' the endpoint : does it work? */

  if (usbfs_clear_halt(fd,EP_INT) < 0)
    {
      perror("usbfs_clear_halt 0x81");
    }

  memset(&act,0,sizeof(act));
  sigemptyset(&act.sa_mask);
  act.sa_flags = SA_RESTART | SA_SIGINFO; 
  act.sa_sigaction = mon_signal;

  if (sigaction(SIGRTMIN,&act,NULL) < 0)
    {
      perror("sigaction");
    }

  usbfs_fill_bulk_urb(&gurb_int,EP_INT,gbuf,sizeof(gbuf),SIGRTMIN);

  if (usbfs_submit_urb(fd,&gurb_int) < 0)
    {
      perror("usbfs_submit_urb");
    }

  if (debug)
    printf("&gurb_int = %p\n",&gurb_int);
  /* that's it ? */
}

/* endpoint 0x87 is used for receiving ATM cells */

void handle_endpoint_87(int fdusb,int fdout)
{
	int i, ret;
	struct usbdevfs_urb * l_urb;
	
	if (usbfs_clear_halt(fdusb,EP_DATA_IN) < 0)
		perror("usbfs_clear_halt 0x87");
	
	if (usb_sync)
    {
		
		/* to handle synchronous read, we fork a new process
		   which does an infinite loop, keeping reading on 
		   endpoint EP_DATA_IN 
		*/
		
		if ((child_pid=fork ()) == 0)
		{
			unsigned char lbuf[64*53];
			struct usbdevfs_urb lurb, *p;
			
			for (;;)
			{
				usbfs_fill_bulk_urb(&lurb,EP_DATA_IN,lbuf,sizeof(lbuf),0);

				if (usbfs_submit_urb(fdusb,&lurb) < 0)
				{
					perror("error reading from usb: usbfs_submit_urb");
					break;
				}
				
				/* wait for SIGRTMIN */
				//pause();
				
				if (usbfs_reapurb(fdusb,&p) < 0)
				{
					perror("error reading from usb: usbfs_reapurb");
					break;
				}
				
				if (p != &lurb)
					printf("p = %p, &lurb = %p\n",p,&lurb);
				
				if (p->buffer != lbuf)
					printf("p->buffer = %p, lbuf = %p\n",p->buffer,lbuf);
				
				if (decode_usb_pkt(lbuf,lurb.actual_length) < 0)
					break;
			}
			
			/* this line is reached only if an error occurs in
			   communicating with usb or pppd */
			
			/* first, we kill our father, which is useless */
			
			if (kill(getppid(),SIGINT) < 0)
				perror("error killing parent process");
			
			_exit (0);
		}
    }
	else
    {
		for (i=0;i<URB_NB;i++)
			urb_submitted[i] = 0;
		
		submit_unsubmitted_urb(fdusb);
		
		for (;;)
		{
			/* wait for a URB to complete */
			ret = usbfs_reapurb(fdusb,&l_urb);
			if (ret < 0)
			{
				if (errno == EINTR)
					continue;
				
				perror("usbfs_reapurb");
				break; 
			}
			
			/* handle the completed URB */
			sig_data_in(l_urb);
		}
    }
}

void handle_endpoint_07(int fdusb,int fdin)
{
  /* this part still need a redesign ... */
  int r,n;
  unsigned char * buf;

  if (usbfs_clear_halt(fdusb,EP_DATA_OUT) < 0)
    perror("usbfs_clear_halt 0x07");

  for (;;)
    {
      n = ppp_read(fdin,&buf,&n);
      if (n<=0)
	  break;

      r = aal5_write(fdusb,buf,n);
      if (r < 0)
	break;
    }
}

void usage()
{
  fprintf(stderr,"pppoa2 version $Name: tar_gz_26042001 $\n");
  fprintf(stderr,"usage: pppoa2 [-d] [-f logfile] -vpi val -vci val \n");
  fprintf(stderr,"  -vpi.vci : defined the vpi.vci that your provider is\n");
  fprintf(stderr,"             using, 8.35 for France, 0.38 for UK\n");
  fprintf(stderr,"  --help : this message\n");
  exit (-1);
}

int main(int argc, char *argv[])
{
  char default_logfile[100];
  char * logfile = default_logfile;
  int fdin, fdout, log, fdusb;
  int i;

  for (i=1;i<argc;i++)
    {
      if (strcmp(argv[i],"-d")==0)
	debug = 1;
      else if (strcmp(argv[i],"-f")==0 && i+1<argc)
	logfile = argv[++i];
      else if (strcmp(argv[i],"-vpi")==0 && i+1<argc)
	my_vpi = atoi(argv[++i]);
      else if (strcmp(argv[i],"-vci")==0 && i+1<argc)
	my_vci = atoi(argv[++i]);
      else if (strcmp(argv[i],"--help")==0)
	usage();
      else
	usage();
    }

  if (my_vpi == -1 || my_vci == -1)
    usage();

  fprintf(stderr,"using vpi.vci=%d.%d\n",my_vpi,my_vci);

  fdin = dup(0);
  fdout = dup(1);

  sprintf(default_logfile,"/tmp/pppoa2.%d.log",getpid());
  log = open(logfile,O_CREAT|O_RDWR|O_TRUNC,0666);
  if (log < 0)
    {
      perror(logfile);
      return -1;
    }

  close(0);
  close(1);
  close(2);

  dup(log);
  dup(log);
  dup(log);

  /* no buffering on stdout & stderr : it doesn't seem to work */

  setbuf(stdout,NULL);
  setbuf(stderr,NULL);

  if (debug)
    printf("fdin=%d, fdout=%d, log=%d\n",fdin,fdout,log);
  close(log);

  /* increase priority */
  if (setpriority(PRIO_PROCESS,getpid(),-20) < 0)
    perror("setpriority");

  /* install HDLC line discipline on our master tty */
  {
    int disc = N_HDLC;
    if (ioctl(fdin,TIOCSETD,&disc) < 0)
      perror("ioctl");
  }

  /* we search for the first USB device matching ST_VENDOR & ST_PRODUCT.
     usbdevfs must be mount on /proc/bus/usb (or you may change the path
     here, according to your config */

  fdusb = usbfs_search("/proc/bus/usb",ST_VENDOR,ST_PRODUCT);
  if (fdusb < 0)
    {
      printf("Where is this crappy modem? huh!\n");
      return -1;
    }

  if (debug)
    printf("Gotcha! fdusb=%d\n",fdusb);

  /* initialize global variables */

  gfdusb = fdusb;
  gfdout = fdout;

  /* initialize our buffer pool, used for writing to USB bus */

  buffer_pool_init(&urb_pool,100,sizeof(struct usbdevfs_urb));
  buffer_pool_init(&buf_pool,100,1367*53);

  /* we claim interface 0, where endpoint 0x81 is */
  /*
  if (usbfs_claimintf(fdusb,0) < 0)
    {
      perror("usbfs_claimintf");
      return -1;
    }
  */

  /* we claim interface 1, where endpoints 0x07 & 0x87 are */
  if (usbfs_claimintf(fdusb,1) < 0)
    {
      perror("usbfs_claimintf");
      return -1;
    }

  /* endpoint 0x81 is used for receiving interrupt transfers */
  //handle_endpoint_81(fdusb);

  /* endpoint 0x87 is used for receiving ATM cells, this function will
   fork a new process that will read ATM cells from the usb and send them
   to pppd */
  handle_endpoint_87(fdusb,fdout);

  /* endpoint 0x07 is used for sending ATM cells, this function
     is an infinite loop and will return only on errors. */
  handle_endpoint_07(fdusb,fdin);

  /* we kill our child pid : this is to be sure that
	 the USB device file will be available for other programs
  */

  if (kill(child_pid,SIGINT) < 0)
	  perror("error killing child process");

  /* we release all the interface we'd claim before exiting */

  //usbfs_releaseintf(fdusb,0);
  usbfs_releaseintf(fdusb,1);

  return 0;
}
