/*
	Author   : Benoit PAPILLAULT <benoit.papillault@free.fr>
	Creation : 05/03/2001

	Use usbdevfs directly (without libusb, which seems to be a crap).
	Take a look at /usr/src/linux/drivers/usb/devio.c for further
	information on IOCTL handle by USB device filesystem.

07/03/2001: it works with 'modprobe usb-uhci' but not with 'modprobe uhci'.
	In this latter case, the error EPIPE is returned from the first
	ioctl.

16/03/2001: try to use libusb, reading 'jukebox' application as an example
        the configuration value used in usb_set_configuration()
	match the field bConfigurationValue in
	URB_FUNCTION_SELECT_CONFIGURATION decoded with Snoopy.

	+ URB_FUNCTION_SELECT_INTERFACE
	  InterfaceNumber = 1 & AlternateSetting = 1.
	  => usb_claim_interface(1)
	   + usb_set_altinterface(1).

	+ URB_FUNCTION_GET_DESCRIPTOR_FROM_DEVICE
	  Index = 3 (iSerialNumber) & DescriptorType = 3 (String)
	  => bmRequestType=0x80, bRequest=6, wValue = 0x0303, wLength=26
	  => usb_control_msg(,0x80,6,0x0303,0,&buf,sizeof(buf),CTRL_TIMEOUT);

17/03/2001: Windows driver initialize configuration 1, interface 1
         & alternate settings 1. Next, experimentally endpoint 5 accept
	 writing. This endpoint is noly available in configuration 1,
	 interface 2, default setting.

	 +  URB_FUNCTION_CLEAR_FEATURE_TO_ENDPOINT
	 => bmRequestType=2, bRequest=1, wValue=0, wIndex=85, wLength=0
	 => usb_control_msg(,2,1,0,0x85,NULL,0,CTRL_TIMEOUT)
*/
	


#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <sys/ioctl.h>

#include <linux/ioctl.h>
#include <linux/types.h>
//#include <linux/usb.h>
#include <linux/usbdevice_fs.h>

#include <usb.h>

#include "urb12.h"
#include "urb13.h"
#include "urb14.h"
#include "urb15.h"
#include "urb16.h"

void breakpoint()
{
}

/* Copy/paste from kernel source, since it's protected by #idef __KERNEL__ */

/* Device descriptor */

/*
struct usb_device_descriptor {
        __u8  bLength;
        __u8  bDescriptorType;
        __u16 bcdUSB;
        __u8  bDeviceClass;
        __u8  bDeviceSubClass;
        __u8  bDeviceProtocol;
        __u8  bMaxPacketSize0;
        __u16 idVendor;
        __u16 idProduct;
        __u16 bcdDevice;
        __u8  iManufacturer;
        __u8  iProduct;
        __u8  iSerialNumber;
        __u8  bNumConfigurations;
} __attribute__ ((packed));
*/
/* Alcatel ADSL SpeedTouch USB modem idVendor/idProduct */

#define ST_VENDOR  0x06b9
#define ST_PRODUCT 0x4061
#define ST_CONFIG 0
#define ST_INTERFACE 1

#define CTRL_TIMEOUT 3000
#define DATA_TIMEOUT 5000

void print_char(unsigned char c)
{
  if (c>=' ' && c<0x7f)
    printf("%c",c);
  /*
    else if (c=='\r' || c=='\n' || c=='\t' || c=='\b')
    printf("%c",c);
  */
  else	printf(".");
}

void dump(unsigned char *buf,int len)
{
  int i,j;
  
  for (i=0;i<len;i+=16)
    {
      for (j=i;j<len && j<i+16;j++)
	printf("%02x ",buf[j]);
      for (;j<i+16;j++)
	printf("   ");
      for (j=i;j<len && j<i+16;j++)
	print_char(buf[j]);
      printf("\n");
    }
}

/* st_init : try to reproduce URB sent by the windows driver
   for this modem, URB got using Snoopy
*/

void st_init(int fd)
{
  struct usbdevfs_ctrltransfer ctrl;
  struct usb_device_descriptor desc;
  int r;

  printf("OK! I found the Alcatel ADSL SpeedTouch USB modem.\n");

  /* URB 1 : GET_DESCRIPTOR_FROM_DEVICE */

  ctrl.requesttype = 0x80;
  ctrl.request     = 6;
  ctrl.value       = 1;
  ctrl.index       = 0;
  ctrl.length      = sizeof(desc);
  ctrl.timeout     = 100; // EXPERIMENTAL
  ctrl.data        = &desc;

  memset(&desc,0,sizeof(desc));

  //  desc.bLength     = sizeof(desc);

  if ((r=ioctl(fd,USBDEVFS_CONTROL,&ctrl)) < 0)
    perror("USBDEVFS_CONTROL");

  printf("r=%d\n",r);
  printf("desc.bLength=%d\n",desc.bLength);

  /* we always get error EPIPE */
}

/* try to open the file and get USB device information,
   if we've found the modem, call st_init()
*/

void test_file(const char *path)
{
	int fd;
	struct usb_device_descriptor desc;

	fd = open(path,O_RDWR);
	if (fd == -1)
	{
		perror(path);
		return ;
	}

	if (read(fd,&desc,sizeof(desc)) == sizeof(desc))
	{
		/* great, we read something */

	  /* check, it match the correct structure */

	  if (desc.bLength == sizeof(desc))
	    {

		printf("=== %s ===\n",path);
		printf("  bLength         = %u\n",desc.bLength);
		printf("  bDescriptorType = %u\n",desc.bDescriptorType);
		printf("  bcdUSB          = %04x\n",desc.bcdUSB);
		printf("  idVendor        = %04x\n",desc.idVendor);
		printf("  idProduct       = %04x\n",desc.idProduct);
		printf("  bcdDevice       = %04x\n",desc.bcdDevice);

		if (desc.idVendor == ST_VENDOR
		    && desc.idProduct == ST_PRODUCT)
		  {
		    st_init(fd);
		  }
	    }
	}

	close(fd);
}

void scan_dir(const char *path)
{
	DIR * dir;
	struct dirent * dirp;

	dir = opendir(path);
	if (dir == NULL)
	{
		perror(path);
		return ;
	}

	while ((dirp=readdir(dir)) != NULL)
	{
		char file[PATH_MAX+1];
		struct stat statbuf;

		if (strcmp(dirp->d_name,".")==0)
			continue;
		if (strcmp(dirp->d_name,"..")==0)
			continue;

		sprintf(file,"%s/%s",path,dirp->d_name);

		if (stat(file,&statbuf)!=0)
		{
			perror(file);
			continue;
		}

		if (S_ISDIR(statbuf.st_mode))
		{
			scan_dir(file);
			continue;
		}

		if (S_ISREG(statbuf.st_mode))
			test_file(file);
	}

	closedir(dir);
}

unsigned char data1[] = {
  0x88, 0xf8, 0x00, 0x00, 0x00, 0x00, 0xf2, 0x01, 
  0x0d, 0x00, 0x00, 0xea, 0x0c, 0x00, 0x00, 0xea, 
  0x0b, 0x00, 0x00, 0xea, 0x0a, 0x00, 0x00, 0xea, 
  0x09, 0x00, 0x00, 0xea, 0x08, 0x00, 0x00, 0xea, 
  0xa4, 0x00, 0x00, 0xea, 0x06, 0x00, 0x00, 0xea, 
  0x32, 0x11, 0xa0, 0xe3, 0x5c, 0x03, 0x9f, 0xe5, 
  0x00, 0x00, 0x81, 0xe5, 0x58, 0x13, 0x9f, 0xe5, 
  0x00, 0x00, 0xa0, 0xe3, 0x00, 0x00, 0x81, 0xe5, 
  0xf8, 0xff, 0xff, 0xea, 0x01, 0x00, 0xa0, 0xe3, 
  0x48, 0x13, 0x9f, 0xe5, 0x00, 0x00, 0x81, 0xe5, 
  0x3c, 0x13, 0x9f, 0xe5, 0x04, 0x00, 0xa0, 0xe3, 
  0x00, 0x00, 0x81, 0xe5, 0xd1, 0x00, 0xa0, 0xe3, 
  0x00, 0xf0, 0x29, 0xe1, 0x01, 0xd9, 0xa0, 0xe3, 
  0xd2, 0x00, 0xa0, 0xe3, 0x00, 0xf0, 0x29, 0xe1, 
  0x24, 0xd3, 0x9f, 0xe5, 0xd3, 0x00, 0xa0, 0xe3, 
  0x00, 0xf0, 0x29, 0xe1, 0x1c, 0xd3, 0x9f, 0xe5, 
  0x00, 0x00, 0xa0, 0xe3, 0x00, 0x10, 0xa0, 0xe3, 
  0xb0, 0x02, 0x9f, 0xe5, 0x08, 0x00, 0x80, 0xe2, 
  0xbc, 0x12, 0x9f, 0xe5, 0x00, 0x10, 0x80, 0xe5, 
  0x5f, 0x10, 0xe0, 0xe3, 0x5d, 0x30, 0xe0, 0xe3, 
  0x5c, 0x50, 0xe0, 0xe3, 0x5b, 0x70, 0xe0, 0xe3, 
  0xac, 0x02, 0x9f, 0xe5, 0xa4, 0x22, 0x9f, 0xe5, 
  0xa8, 0x42, 0x9f, 0xe5, 0xa8, 0x62, 0x9f, 0xe5, 
  0x00, 0x10, 0x80, 0xe5, 0x00, 0x30, 0x82, 0xe5, 
  0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
  0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
  0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
  0x00, 0x50, 0x84, 0xe5, 0x00, 0x50, 0x84, 0xe5, 
  0x00, 0x70, 0x86, 0xe5, 0x54, 0x02, 0x9f, 0xe5, 
  0x08, 0x00, 0x80, 0xe2, 0x5c, 0x12, 0x9f, 0xe5, 
  0x00, 0x10, 0x80, 0xe5, 0x6c, 0x02, 0x9f, 0xe5, 
  0x6c, 0x12, 0x9f, 0xe5, 0x00, 0x10, 0x80, 0xe5, 
  0x00, 0x10, 0xa0, 0xe3, 0x98, 0x02, 0x9f, 0xe5, 
  0x00, 0x10, 0x80, 0xe5, 0x94, 0x02, 0x9f, 0xe5, 
  0x00, 0x10, 0x80, 0xe5, 0x01, 0x09, 0xa0, 0xe3, 
  0x00, 0x10, 0xa0, 0xe3, 0x02, 0x26, 0xa0, 0xe3, 
  0x04, 0x30, 0xa0, 0xe3, 0x03, 0x10, 0x80, 0xe6, 
  0x02, 0x00, 0x50, 0xe1, 0xfc, 0xff, 0xff, 0x1a, 
  0x7f, 0x19, 0xa0, 0xe3, 0x1e, 0x30, 0xa0, 0xe3, 
  0x03, 0x3b, 0x83, 0xe2, 0x00, 0x00, 0xa0, 0xe3, 
  0x00, 0x2a, 0x83, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
  0x14, 0x00, 0x50, 0xe3, 0x04, 0x20, 0x81, 0xe4, 
  0xfa, 0xff, 0xff, 0xba, 0x02, 0xe0, 0xa0, 0xe3, 
  0x03, 0xeb, 0x8e, 0xe2, 0x0e, 0x20, 0xa0, 0xe1, 
  0x18, 0x00, 0x50, 0xe3, 0x04, 0x00, 0x00, 0xaa, 
  0x00, 0xca, 0x82, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
  0x18, 0x00, 0x50, 0xe3, 0x04, 0xc0, 0x81, 0xe4, 
  0xfa, 0xff, 0xff, 0xba, 0x40, 0x00, 0x50, 0xe3, 
  0x04, 0x00, 0x00, 0xaa, 0x00, 0x2a, 0x83, 0xe1, 
  0x01, 0x00, 0x80, 0xe2, 0x40, 0x00, 0x50, 0xe3, 
  0x04, 0x20, 0x81, 0xe4, 0xfa, 0xff, 0xff, 0xba, 
  0x01, 0x0a, 0x50, 0xe3, 0x04, 0x00, 0x00, 0xaa, 
  0x00, 0x2a, 0x8e, 0xe1, 0x01, 0x00, 0x80, 0xe2, 
  0x01, 0x0a, 0x50, 0xe3, 0x04, 0x20, 0x81, 0xe4, 
  0xfa, 0xff, 0xff, 0xba, 0x7f, 0x09, 0xa0, 0xe3, 
  0x03, 0x40, 0xa0, 0xe3, 0x10, 0x0f, 0x02, 0xee, 
  0x10, 0x4f, 0x03, 0xee, 0x10, 0x0f, 0x07, 0xee, 
  0x10, 0x0f, 0x05, 0xee, 0x3d, 0x00, 0xa0, 0xe3, 
  0x10, 0x0f, 0x01, 0xee, 0x01, 0x73, 0xa0, 0xe3, 
  0x04, 0x00, 0xa0, 0xe3, 0x05, 0x00, 0xc7, 0xe5, 
  0x01, 0x00, 0xa0, 0xe3, 0x06, 0x00, 0xc7, 0xe5, 
  0x01, 0x00, 0xa0, 0xe3, 0x07, 0x00, 0xc7, 0xe5, 
  0x03, 0x00, 0xa0, 0xe3, 0x08, 0x00, 0xc7, 0xe5, 
  0x00, 0x00, 0x40, 0x01, 0x12, 0x88, 0xd4, 0xf2, 
  0x01, 0x00, 0x00, 0xce, 0x01, 0xa0, 0xe3, 0x09, 
  0x00, 0xc7, 0xe5, 0x80, 0x00, 0xa0, 0xe3, 0x01, 
  0x00, 0xc7, 0xe5, 0x53, 0x00, 0xa0, 0xe3, 0x00, 
  0xf0, 0x29, 0xe1, 0x94, 0x11, 0x9f, 0xe5, 0x21, 
  0x03, 0xa0, 0xe3, 0x00, 0x10, 0x80, 0xe5, 0x10, 
  0x10, 0xa0, 0xe3, 0x22, 0x01, 0xa0, 0xe3, 0x00, 
  0x10, 0x80, 0xe5, 0x04, 0x00, 0x8f, 0xe2, 0x01, 
  0x00, 0x80, 0xe2, 0x10, 0xff, 0x2f, 0xe1, 0x5e, 
  0x4f, 0xb8, 0x78, 0x40, 0x1e, 0xfc, 0xd1, 0x38, 
  0x78, 0xf9, 0x7a, 0x09, 0x02, 0xbc, 0x7a, 0x21, 
  0x43, 0x09, 0x02, 0x7c, 0x7a, 0x21, 0x43, 0x09, 
  0x02, 0x3c, 0x7a, 0x21, 0x43, 0x11, 0x38, 0x01, 
  0x28, 0x2d, 0xdb, 0x03, 0x28, 0x2b, 0xdc, 0x80, 
  0x00, 0x55, 0x4c, 0x24, 0x58, 0xa7, 0x46, 0x46, 
  0x4b, 0x0e, 0x33, 0x7a, 0x7b, 0x12, 0x02, 0x3c, 
  0x7b, 0x22, 0x43, 0x12, 0xd0, 0x14, 0x1c, 0xf4, 
  0x3c, 0xff, 0x3c, 0x0e, 0xd5, 0x40, 0x07, 0x02, 
  0xd0, 0x1c, 0x1c, 0x0b, 0x1c, 0x21, 0x1c, 0x52, 
  0x1e, 0x8c, 0x5c, 0x9c, 0x54, 0xfb, 0xd1, 0x3b, 
  0xa4, 0x14, 0x23, 0x23, 0x60, 0xc0, 0x0f, 0x64, 
  0x30, 0x00, 0xe0, 0x66, 0x20, 0x78, 0x70, 0xcb, 
  0xe7, 0x01, 0xa2, 0x10, 0x47, 0x00, 0x00, 0x01, 
  0xf0, 0xa0, 0xe1, 0x04, 0x20, 0x8f, 0xe2, 0x01, 
  0x20, 0x82, 0xe2, 0x12, 0xff, 0x2f, 0xe1, 0x60, 
  0x20, 0xf0, 0xe7, 0x04, 0xe0, 0x4e, 0xe2, 0x00, 
  0x40, 0x2d, 0xe9, 0x00, 0xe0, 0x4f, 0xe1, 0x07, 
  0x40, 0x2d, 0xe9, 0xa8, 0x10, 0x8f, 0xe2, 0x00, 
  0x00, 0x91, 0xe5, 0x01, 0x00, 0x20, 0xe2, 0x00, 
  0x00, 0x81, 0xe5, 0x00, 0x00, 0x50, 0xe3, 0x03, 
  0x00, 0x00, 0x0a, 0x0d, 0x10, 0xa0, 0xe3, 0xcc, 
  0x20, 0x9f, 0xe5, 0x00, 0x10, 0x82, 0xe5, 0x02, 
  0x00, 0x00, 0xea, 0x0f, 0x10, 0xa0, 0xe3, 0xbc, 
  0x20, 0x9f, 0xe5, 0x00, 0x10, 0x82, 0xe5, 0x01, 
  0x00, 0xa0, 0xe3, 0x90, 0x10, 0x9f, 0xe5, 0x00, 
  0x00, 0x81, 0xe5, 0x6c, 0x10, 0x8f, 0xe2, 0x00, 
  0x00, 0x91, 0xe5, 0x01, 0x00, 0x40, 0xe2, 0x00, 
  0x00, 0x81, 0xe5, 0x00, 0x00, 0x50, 0xe3, 0x02, 
  0x00, 0x00, 0x1a, 0x94, 0x00, 0x9f, 0xe5, 0x94, 
  0x10, 0x9f, 0xe5, 0x00, 0x00, 0x81, 0xe5, 0x90, 
  0x20, 0x9f, 0xe5, 0x00, 0x10, 0x92, 0xe5, 0x07, 
  0x40, 0xbd, 0xe8, 0x0e, 0xf0, 0x69, 0xe1, 0x00, 
  0x80, 0xfd, 0xe8, 0x00, 0x00, 0x00, 0x8c, 0x20, 
  0x00, 0x00, 0x88, 0x50, 0x00, 0x00, 0x01, 0x02, 
  0x00, 0x00, 0x00, 0xf4, 0x00, 0x00, 0x00, 0x39, 
  0x00, 0x00, 0x00, 0x10, 0x40, 0x08, 0x00, 0x00, 
  0x40, 0x00, 0x00, 0x20, 0x40, 0x00, 0x00, 0x30, 
  0x4e, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa0, 0x15, 
  0x80, 0x00, 0x00, 0x15, 0x2f, 0x8c, 0x9b, 0x00, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x5d, 
  0x02, 0x00, 0x00, 0x5d, 0x02, 0x00, 0x00, 0x97, 
  0x02, 0x00, 0x00, 0xff, 0xff, 0x00, 0x00, 0x40, 
  0x00, 0x40, 0x01, 0x64, 0x00, 0x50, 0x01, 0x70, 
  0x3e, 0x00, 0x00, 0xe0, 0x3c, 0x00, 0x00, 0x20, 
  0x00, 0x00, 0x90, 0x20, 0x00, 0x00, 0xa4, 0x85, 
  0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x78, 
  0x03, 0x00, 0x00, 0x58, 0x00, 0x50, 0x01, 0x70, 
  0xc3, 0x00, 0x00, 0x60, 0x00, 0x50, 0x01, 0x04, 
  0x00, 0x00, 0x84, 0x40, 0x01, 0x12, 0x08, 0x04, 
  0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x14
};

struct usb_dev_handle * st_handle;

void read_interrupts()
{
  char buf[100];
  int n;

  for (;;)
    {
      /*printf("reading interrupt ...\n");*/
      n = usb_bulk_read(st_handle,0x81,buf,10,10);
      if (n<0)
	/*printf("usb_bulk_read: %s\n",usb_strerror())*/;
      else
	{
	  printf("interrupt read : ");
	  dump(buf,n);
	}

      usleep(50 * 1000);
    }
}

int main(int argc, char *argv[])
{
  struct usb_bus * bus;
  struct usb_device * dev;
  struct usb_device * st_dev = NULL;
  FILE * microcode;
  char val = 0;
  pid_t pid;

  /*
    You should have insmod'd "usbcore" & "usb-uhci" according to
    your USB host controller. At this stage, both green leds should be ON
  */

  int i,n, idx=0;
  char buf[0x1000];

  usb_init();
  usb_find_busses();
  usb_find_devices();

  for (bus=usb_busses;bus;bus=bus->next)
    for (dev=bus->devices;dev;dev=dev->next)
      if (dev->descriptor.idVendor == ST_VENDOR
	  && dev->descriptor.idProduct == ST_PRODUCT)
	st_dev = dev;

  if (st_dev == NULL)
    {
      printf("Can't find ADSL modem!\n");
      return -1;
    }

  printf("Found the ADSL modem with VendorID=%04x & ProductID=%04x\n",
	 st_dev->descriptor.idVendor,st_dev->descriptor.idProduct);

  st_handle = usb_open(st_dev);
  if (st_handle == NULL)
    {
      printf("Couldn't get device handle for speedtouch\n");
      return -1;
    }

  if (usb_set_configuration(st_handle,1 /*ST_CONFIG*/) < 0)
    {
      printf("Unable to set configuration [%s]\n",usb_strerror());
      return -1;
    }

  if (usb_claim_interface(st_handle,1 /*ST_INTERFACE*/) <0)
    {
      printf("Unable to set interface\n");
      return -1;
    }

  if (usb_set_altinterface(st_handle,1) < 0)
    {
      printf("usb_set_altinterface:%s\n",usb_strerror());
      return -1;
    }

  n=usb_control_msg(st_handle,0x80,6,0x300|st_dev->descriptor.iManufacturer,0,
		    (char *)&buf,sizeof(buf),CTRL_TIMEOUT);
  if (n > 0)
    {
      printf("Manufacturer  : ");
      for (i=2;i<n;i+=2)
	printf("%c",buf[i]);
      printf("\n");
    }

  n=usb_control_msg(st_handle,0x80,6,0x300|st_dev->descriptor.iProduct,0,
		    (char *)&buf,sizeof(buf),CTRL_TIMEOUT);
  if (n > 0)
    {
      printf("Product       : ");
      for (i=2;i<n;i+=2)
	printf("%c",buf[i]);
      printf("\n");
    }
  
  n=usb_control_msg(st_handle,0x80,6,0x300|st_dev->descriptor.iSerialNumber,0,
		    (char *)&buf,sizeof(buf),CTRL_TIMEOUT);
  if (n > 0)
    {
      printf("Serial number : ");
      for (i=2;i<n;i+=2)
	printf("%c",buf[i]);
      printf("\n");
    }

  /* PipeHandle 0x8609C420 => endpoint 5 */
  
  /* URB 8 */
  n = usb_bulk_write(st_handle,0x5,data1,sizeof(data1),DATA_TIMEOUT);
  if (n < 0)
    printf("usb_bulk_write: %s\n",usb_strerror());

  /* Just after usb_bulk_write, ADSL led turns off, USB leds blinks green */

  /* URB 9 */
  n = usb_control_msg(st_handle,
		      USB_RECIP_ENDPOINT /*2*/,
		      USB_REQ_CLEAR_FEATURE /*1*/,
		      0,0x85,NULL,0,CTRL_TIMEOUT);
  if (n < 0)
    printf("CLEAR_FEATURE: %s\n",usb_strerror());

  n = usb_resetep(st_handle,0x85); /* urb 10, same ep as urb 11 */
  if (n < 0)
    printf("usb_resetep: %s\n",usb_strerror());

  /* URB 11 */
  n = usb_bulk_read(st_handle,0x85,buf,0x1ff /* !!! */,DATA_TIMEOUT);
  if (n < 0)
    printf("usb_bulk_read: %s\n",usb_strerror());

  /*
    printf("n=%d\n",n);
    dump(buf,n);
  */

  /* URB 12 to 139 */

  microcode = fopen("microcode.bin","r");
  if (microcode != NULL)
    {
      unsigned char mm[0x1000];
      int len;

      while ((len = fread(mm,1,sizeof(mm),microcode)) > 0)
	{
	  if (len < 0x1000)
	    breakpoint();
	  n = usb_bulk_write(st_handle,0x05,mm,len,DATA_TIMEOUT);
	  if (n < 0)
	    printf("usb_bulk_write: %s\n",usb_strerror());
	  else
	    /*printf("uploading %d bytes microcode... ok.\n",len)*/;
	}

      fclose (microcode);
      printf("microcode downloaded sucessfully!\n");
      /* before downloading microcode :
	 + USB  led : blinking
	 + ADSL led : off

	 after downloading microcode :
	 + USB  led : green
	 + ADSL led : red/green
      */
    }
  
  /* URB 140 */
  n = usb_control_msg(st_handle,
		      USB_RECIP_ENDPOINT /*2*/,
		      USB_REQ_CLEAR_FEATURE /*1*/,
		      0,0x85,NULL,0,CTRL_TIMEOUT);
  if (n < 0)
    printf("CLEAR_FEATURE: %s\n",usb_strerror());

  /* URB 141 */
  n = usb_resetep(st_handle,0x85); /* urb 10, same ep as urb 11 */
  if (n < 0)
    printf("usb_resetep: %s\n",usb_strerror());

  /* URB 142 */
  n = usb_bulk_read(st_handle,0x85,buf,0x1ff /* !!! */,DATA_TIMEOUT);
  if (n < 0)
    printf("usb_bulk_read: %s\n",usb_strerror());

  /*dump(buf,n);*/

  /* URB 143 */
  sleep(1);

  {
    unsigned char c = 1;
    n = usb_control_msg(st_handle,0xc0,0x12,0x07,0x00,&c,sizeof(c),100);
    if (n < 0)
      printf("usb_control_msg: %s\n",usb_strerror());
    printf("c=0x%x\n",c);
  }

  /* URB 144 */
  if (usb_claim_interface(st_handle,1) < 0)
    {
      printf("usb_claim_interface: %s\n",usb_strerror());
      return -1;
    }

  if (usb_set_altinterface(st_handle,2) < 0)
    {
      printf("usb_set_altinterface: %s\n",usb_strerror());
      return -1;
    }

  /* URB 145 */
  if (usb_claim_interface(st_handle,1) < 0)
    {
      printf("usb_claim_interface: %s\n",usb_strerror());
      return -1;
    }

  if (usb_set_altinterface(st_handle,1) < 0)
    {
      printf("usb_set_altinterface: %s\n",usb_strerror());
      return -1;
    }
  
  /* URB 147 */
  buf[0] = 0x1c; buf[1] = 0x50;
  n = usb_control_msg(st_handle,0x40,0x01,0xb,0,buf,2,CTRL_TIMEOUT);
  if (n<0)
    printf("usb_control_msg: %s\n",usb_strerror());

  /* URB 148 */
  buf[0] = 0x32; buf[1] = 00;
  n = usb_control_msg(st_handle,0x40,0x01,0x02,0x00,buf,2,CTRL_TIMEOUT);
  if (n<0)
    printf("usb_control_msg: %s\n",usb_strerror());

  /* URB 149 */
  buf[0] = 1; buf[1] = 0; buf[2] = 1;
  n = usb_control_msg(st_handle,0x40,1,3,0,buf,3,CTRL_TIMEOUT);
  if (n<0)
    printf("usb_control_msg: %s\n",usb_strerror());

  /* URB 150 */
  buf[0] = 1; buf[1] = 0; buf[2] = 1;
  n = usb_control_msg(st_handle,0x40,1,4,0,buf,3,CTRL_TIMEOUT);
  if (n<0)
    printf("usb_control_msg: %s\n",usb_strerror());

  /* URB 151 : this buffer match (if you reverse one byte with its neighbour)
   the label on the manta box!!! in my case : 3EC18607CAAB06 */

  n = usb_control_msg(st_handle,0xc0,0x26,0xf7,0,buf,0xe,CTRL_TIMEOUT);
  if (n<0)
    printf("usb_control_msg: %s\n",usb_strerror());

  printf("Reference (see undex your box) : "); dump(buf,n);

  /* URB 152 : match the serial number in a UNICODE string descriptor */
  n = usb_control_msg(st_handle,0xc0,0x26,0x6b,0,buf,0x1a,CTRL_TIMEOUT);
  if (n<0)
    printf("usb_control_msg: %s\n",usb_strerror());

  /*dump(buf,n);*/

  /* URB 153 */
  n = usb_control_msg(st_handle,2,1,0,0x87,NULL,0,CTRL_TIMEOUT);
  if (n<0)
    printf("usb_control_msg: %s\n",usb_strerror());

  /* URB 154 */
  n = usb_resetep(st_handle,0x87);
  if (n<0)
    printf("usb_resetep: %s\n",usb_strerror());

  /* URB 155 - 162 */
  n = usb_bulk_read(st_handle,0x87,buf,0xd40,100);
  if (n<0)
    printf("Like windows crap: %s\n",usb_strerror());

  /* URB 163 */
  n = usb_control_msg(st_handle,0xc0,0x12,4,0,buf,2,CTRL_TIMEOUT);
  if (n<0)
    printf("usb_control_msg: %s\n",usb_strerror());

  if ((pid=fork()) == 0)
    {
      read_interrupts();
      exit (-1);
    }

  for (;;)
    {
      /*printf("---\n");*/
      /* URB 164 */
      n = usb_control_msg(st_handle,0xc0,1,0xe,0,buf,1,CTRL_TIMEOUT);
      if (n<0)
	printf("URB 164: %s\n",usb_strerror());
      if (n!=1)
	{
	  printf("c0 01 0e => "); dump(buf,n);
	}
      
      /* URB 165 */
      n = usb_control_msg(st_handle,0xc0,1,0xf,0,buf,1,CTRL_TIMEOUT);
      if (n<0)
	printf("URB 165: %s\n",usb_strerror());
      if (n != 1)
	{
	  printf("c0 01 0f => "); dump(buf,n);
	}

      /* URB 166 */
      n = usb_control_msg(st_handle,0xc0,0x12,0xd,0,buf,4,CTRL_TIMEOUT);
      if (n<0)
	printf("URB 166: %s\n",usb_strerror());
      if (n != 4)
	{
	  printf("c0 12 0d => "); dump(buf,n);
	}

      /* URB 167 */
      n = usb_control_msg(st_handle,0xc0,0x12,7,0,buf,1,CTRL_TIMEOUT);
      if (n<0)
	printf("URB 167: %s\n",usb_strerror());
      if (n!=1)
	{
	  printf("c0 12 07 => "); dump(buf,n);
	}

      if (val != buf[0])
	{
	  val = buf[0];
	  printf("0x%x\n",val);
	}
      
      if (buf[0] == 0x20)
	{
	  n = usb_control_msg(st_handle,0xc0,0x12,0xb,0,buf,8,CTRL_TIMEOUT);
	  if (n<0)
	    printf("URB 225: %s\n",usb_strerror());
	  printf("c0 12 0b => "); dump(buf,n);

	  break;
	}

    }

  usb_close(st_handle);
  kill(pid,SIGTERM);
  return 0;
}

